@@ -101,6 +101,7 @@ class SAMAP(object):
             If True saves the processed SAM objects corresponding to each species to an `.h5ad` file.
             This argument is unused if preloaded SAM objects are passed in to SAMAP.
         """
+
         for key,data in zip(sams.keys(),sams.values()):
             if not (isinstance(data, str) or isinstance(data, SAM)):
                 raise TypeError(f"Input data {key} must be either a path or a SAM object.")
@@ -219,7 +220,7 @@ class SAMAP(object):
         hom_edge_mode: typing.Optional[str] = "pearson",
         scale_edges_by_corr: typing.Optional[bool] = True,
         neigh_from_keys: typing.Optional[dict] = None,
-        
+        pairwise: typing.Optional[bool] = True
     ):
         """Runs the SAMap algorithm.
 
@@ -272,11 +273,27 @@ class SAMAP(object):
         neigh_from_key2 : bool, optional, default False
             If True, species 2 neighborhoods are calculated directly from the chosen clustering (`self.key2`).
             Cells within the same cluster belong to the same neighborhood.
+
+        pairwise: bool, optional, default True
+            If True, compute mutual nearest neighborhoods independently between each pair of species.
+            If False, compute mutual nearest neighborhoods between each species and all other species.
+            This parameter is ignored if there are only two species to be mapped.
+
+           `pairwise=True` would prevent outgroup species from being unmapped.
+            For example, if `pairwise=False`, when mapping three species like human, mouse, and zebrafish, human and mice
+            will preferentially map to each other, leaving zebrafish unmapped.
+
+            Set `pairwise=False` when you'd like to be able to tell which cell types are more or less similar 
+            between different species. Set `pairwise=True` when you'd like to remove the effects of evolutionary
+            distance. This is particularly useful when using multiple reference species to annotate an unlabeled
+            dataset.
             
         Returns
         -------
         samap - Species-merged SAM object
         """
+        self.pairwise = pairwise
+
         ids = self.ids
         sams = self.sams
         gnnm = self.gnnm
@@ -305,15 +322,16 @@ class SAMAP(object):
             corr_mode=hom_edge_mode,
             scale_edges_by_corr = scale_edges_by_corr,
             neigh_from_keys=neigh_from_keys,
+            pairwise=pairwise
         )
         samap = smap.final_sam
         self.samap = samap
         self.ITER_DATA = smap.ITER_DATA
 
-        print("Alignment score ---", _avg_as(samap).mean())
+        print("Alignment score ---", _avg_as(samap, pairwise=pairwise).mean())
         if umap:
             print("Running UMAP on the stitched manifolds.")
-            sc.tl.umap(self.samap.adata,min_dist=0.1,init_pos='random')
+            sc.tl.umap(self.samap.adata,min_dist=0.1,init_pos='random',maxiter = 500 if self.samap.adata.shape[0] <= 10000 else 200)
         
         
         ix = pd.Series(data = np.arange(samap.adata.shape[1]),index = samap.adata.var_names)[gns].values
@@ -356,8 +374,7 @@ class SAMAP(object):
         print("Running UMAP on the stitched manifolds.")
         ids = self.ids
         sams = self.sams
-        
-        sc.tl.umap(self.samap.adata,min_dist=0.1,init_pos='random')
+        sc.tl.umap(self.samap.adata,min_dist=0.1,init_pos='random', maxiter = 500 if self.samap.adata.shape[0] <= 10000 else 200)
         for sid in ids:
             sams[sid].adata.obsm['X_umap_samap'] = self.samap.adata[sams[sid].adata.obs_names].obsm['X_umap']               
 
@@ -661,7 +678,8 @@ class _Samap_Iter(object):
         return gnnmu
 
     def run(self, NUMITERS=3, NHS=None, K=20, corr_mode='pearson', NCLUSTERS=1,
-                  scale_edges_by_corr=True, THR=0, neigh_from_keys=None, ncpus=os.cpu_count()):
+                  scale_edges_by_corr=True, THR=0, neigh_from_keys=None, pairwise=True,
+                  ncpus=os.cpu_count()):
         sams = self.sams
         gnnm = self.gnnm
         gns_dict = self.gns_dict
@@ -703,6 +721,7 @@ class _Samap_Iter(object):
                 keys=keys,
                 scale_edges_by_corr=scale_edges_by_corr,
                 neigh_from_keys=neigh_from_keys,
+                pairwise=pairwise
             )
 
             self.samap = sam4
@@ -726,7 +745,7 @@ class _Samap_Iter(object):
             print(
                 "ITERATION: " + str(i),
                 "\nAverage alignment score (A.S.): ",
-                _avg_as(sam4).mean(),
+                _avg_as(sam4, pairwise=pairwise).mean(),
                 "\nMax A.S. improvement:",
                 diffmax,
                 "\nMin A.S. improvement:",
@@ -778,6 +797,7 @@ def _mapper(
     keys=None,
     scale_edges_by_corr=False,
     neigh_from_keys=None,
+    pairwise=True,
     **kwargs
 ):
     if NHS is None:
@@ -791,7 +811,7 @@ def _mapper(
             neigh_from_keys[sid] = False    
     
     if mdata is None:
-        mdata = _mapping_window(sams, gnnm, gn, K=K)
+        mdata = _mapping_window(sams, gnnm, gn, K=K, pairwise=pairwise)
 
     if k is None:
         k1 = sams[list(sams.keys())[0]].run_args.get("k", 20)
@@ -901,10 +921,31 @@ def _mapper(
         # change new max scores to old max scores
         D = D.multiply(ma/ma2).tocsr()
     
-    Dk = sparse_knn(D, k1).tocsr()
+    species_list = []
+    for sid in sams.keys():
+        species_list += [sid]*sams[sid].adata.shape[0]    
+    species_list = np.array(species_list)
 
+    if not pairwise or len(sams.keys())==2:
+        Dk = sparse_knn(D, k1).tocsr()
+        denom = k1
+    else:
+        Dk=[]
+        for sid1 in sams.keys():
+            row=[]
+            for sid2 in sams.keys():
+                if sid1 != sid2:
+                    Dsubk = sparse_knn(D[species_list==sid1][:,species_list==sid2], k1).tocsr()
+                else:
+                    Dsubk = sp.sparse.csr_matrix((sams[sid1].adata.shape[0],)*2)
+                row.append(Dsubk)
+            Dk.append(sp.sparse.hstack(row))
+        Dk = sp.sparse.vstack(Dk).tocsr()
+        denom = (k1 * (len(sams.keys())-1))
+            
     sr = Dk.sum(1).A    
-    x = 1 - sr.flatten() / k1
+    
+    x = 1 - sr.flatten() / denom
     
     sr[sr==0]=1
     st = Dk.sum(0).A.flatten()[None,:]
@@ -932,15 +973,10 @@ def _mapper(
     NNM = omp.multiply(x[:, None])
     NNM = (NNM+Dk).tolil()
     NNM.setdiag(0)
-    # """
-
+   
     print("Concatenating SAM objects...")
     sam3 = _concatenate_sam(sams, NNM)
     
-    species_list = []
-    for sid in sams.keys():
-        species_list += [sid]*sams[sid].adata.shape[0]
-    
     sam3.adata.obs["species"] = pd.Categorical(species_list)
 
     sam3.adata.uns["gnnm_corr"] = mdata.get("gnnm_corr",None)
@@ -950,7 +986,7 @@ def _mapper(
 
     if umap:
         print("Computing UMAP projection...")
-        sc.tl.umap(sam3.adata, min_dist=0.1)
+        sc.tl.umap(sam3.adata, min_dist=0.1, maxiter = 500 if sam3.adata.shape[0] <= 10000 else 200)
     return sam3
 
 def _refine_corr(
@@ -1382,14 +1418,16 @@ def _get_pairs(sams, gnnm, gns_dict, NOPs1=0, NOPs2=0):
 
     return B
 
-def _avg_as(s):
+def _avg_as(s,pairwise=True):
     x = q(s.adata.obs['batch'])
     xu = np.unique(x)
     a = []
     for i in range(xu.size):
         a.extend(s.adata.obsp['connectivities'][x==xu[i],:][:,x!=xu[i]].sum(1).A.flatten())
-
-    return  np.array(a) / s.adata.obsp['knn'][0].data.size * (xu.size-1)
+    if pairwise:
+        return  np.array(a) / s.adata.obsp['knn'][0].data.size
+    else:
+        return  np.array(a) / s.adata.obsp['knn'][0].data.size * (xu.size-1)
 
 
 def _parallel_init(iXavg, ipairs, igns_dictO, iT2, iCORR, icorr_mode,icl,ics):
@@ -1722,7 +1760,7 @@ def _united_proj(wpca1, wpca2, k=20, metric="cosine", ef=200, M=48):
 def _tanh_scale(x,scale=10,center=0.5):
     return center + (1-center) * np.tanh(scale * (x - center))
 
-def _mapping_window(sams, gnnm=None, gns=None, K=20):
+def _mapping_window(sams, gnnm=None, gns=None, K=20, pairwise=True):
     k = K
     output_dict = {}
     if gnnm is not None and gns is not None:
@@ -1737,34 +1775,51 @@ def _mapping_window(sams, gnnm=None, gns=None, K=20):
         Ws={}
         ss={}
         As={}
-        species_indexer = []    
+        species_indexer = []   
+        genes_indexer = [] 
         for sid in sams.keys():
             gs[sid] = gns[np.in1d(gns,q(sams[sid].adata.var_names))]
             adatas[sid] = sams[sid].adata[:,gs[sid]]
             Ws[sid] = adatas[sid].var["weights"].values
             ss[sid] = std.fit_transform(adatas[sid].X).multiply(Ws[sid][None,:]).tocsr()
             species_indexer.append(np.arange(ss[sid].shape[0]))
+            genes_indexer.append(np.arange(gs[sid].size))
 
         for i in range(1,len(species_indexer)):
             species_indexer[i] = species_indexer[i]+species_indexer[i-1].max()+1
+            genes_indexer[i] = genes_indexer[i]+genes_indexer[i-1].max()+1
 
-        su = gnnm.sum(0).A
+        su = gnnm_corr.sum(0).A
         su[su==0]=1
         gnnm_corr = gnnm_corr.multiply(1/su).tocsr()
         
-        
-        ttt=time.time()
-        print('Translating feature spaces.')    
+        X = sp.sparse.block_diag(list(ss.values())).tocsr()
         W = np.concatenate(list(Ws.values())).flatten()
 
-        X = sp.sparse.block_diag(list(ss.values())).tocsr()
-        Xtr = []
-        for i,sid in enumerate(sams.keys()):
-            Xtr.append(X[species_indexer[i]].dot(gnnm_corr))
-            Xtr[-1] = std.fit_transform(Xtr[-1]).multiply(W[None,:])
-        Xtr = sp.sparse.vstack(Xtr)
-        Xc = X + Xtr    
-        
+        ttt=time.time()
+        if pairwise:
+            print('Translating feature spaces pairwise.')
+            Xtr = []
+            for i,sid1 in enumerate(sams.keys()):
+                xtr = []
+                for j,sid2 in enumerate(sams.keys()):
+                    if i != j:
+                        xtr.append(X[species_indexer[i]][:,genes_indexer[i]].dot(gnnm_corr[genes_indexer[i]][:,genes_indexer[j]]))
+                        xtr[-1] = std.fit_transform(xtr[-1]).multiply(W[genes_indexer[j]][None,:])
+                    else:
+                        xtr.append(sp.sparse.csr_matrix((species_indexer[i].size,genes_indexer[i].size)))
+                Xtr.append(sp.sparse.hstack(xtr))
+            Xtr = sp.sparse.vstack(Xtr)
+        else:
+            print('Translating feature spaces all-to-all.')    
+
+            Xtr = []
+            for i,sid in enumerate(sams.keys()):
+                Xtr.append(X[species_indexer[i]].dot(gnnm_corr))
+                Xtr[-1] = std.fit_transform(Xtr[-1]).multiply(W[None,:])
+            Xtr = sp.sparse.vstack(Xtr)
+        Xc = (X + Xtr).tocsr()
+
         mus = []        
         for i,sid in enumerate(sams.keys()):
             mus.append(Xc[species_indexer[i]].mean(0).A.flatten())
